# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass
from itertools import product
from math import prod

from qat.purr.compiler.builders import QuantumInstructionBuilder
from qat.purr.compiler.instructions import DeviceUpdate, Instruction, Sweep, Variable
from qat.purr.utils.logger import get_default_logger

logger = get_default_logger()


@dataclass
class VariableAccessor:
    """Contains the information needed to access a variable instance in an instruction
    builder."""

    instruction_index: int
    attribute: str
    variable_name: str

    @classmethod
    def create_accessors(
        cls, instruction: Instruction, index: int, names: list[str]
    ) -> list["VariableAccessor"]:
        """Creates a list of VariableAccessors for the given instruction, looking for
        variables with the given names."""

        accessors = []
        for key, val in instruction.__dict__.items():
            if key.startswith("__"):
                continue
            if isinstance(val, Variable) and val.name in names:
                accessors.append(cls(index, key, val.name))
        return accessors


class DeviceAssignSet:
    """Contains a set of device assignments to be applied to a builder."""

    def __init__(self, assigns: list[DeviceUpdate]):
        self.assigns = assigns

    @contextmanager
    def apply(self):
        """Temporarily applies the device assignments to the targets.

        .. warning::

            This will mutate the components of the hardware model, potentially invalidating
            any analysis already done on instantiation of pipelines. Futhermore, since this
            is used in conjuntion with sweeping, it is not thread safe. This is only
            intended to be used in the SweepPipeline, and should not be used elsewhere.
            It is recommended that this is used with a hardware model, or hardware model
            loader that is not shared with any other pipeline.
        """

        original_values = [self._get_attribute(assign) for assign in self.assigns]

        try:
            for assign in self.assigns:
                self._set_attribute(assign, assign.value)
            yield
        finally:
            for assign, original in zip(self.assigns, original_values):
                self._set_attribute(assign, original)

    def _get_attribute(self, assign: DeviceUpdate):
        if isinstance(assign.target, dict):
            return assign.target[assign.attribute]
        else:
            return getattr(assign.target, assign.attribute)

    def _set_attribute(self, assign: DeviceUpdate, value):
        if isinstance(assign.target, dict):
            assign.target[assign.attribute] = value
        else:
            setattr(assign.target, assign.attribute, value)

    def __len__(self):
        return len(self.assigns)


@dataclass
class SweepInstance:
    """Contains a single instance of a sweep, with the variable names and values."""

    variables: dict[str, complex | float]
    builder: QuantumInstructionBuilder
    device_assigns: DeviceAssignSet

    @property
    def has_device_assigns(self) -> bool:
        return len(self.device_assigns) > 0


class SweepFlattener:
    """Extracts sweeps and device assigns from a builder, and acts as a factory for creating
    a set of new builder that is free of sweeps and device assigns.

    This can be used as an iterator to return a series of builders with the sweep values
    injected into the relevant instructions. This works by first extracting sweep
    instructions, and then locating all variables within instructions that match the sweep
    variable names. It only searches a single depth (e.g. within the top-level dictionary
    describing the instruction). This may be adjusted in the future if we need to perform
    a deeper search.
    """

    def __init__(self, builder: QuantumInstructionBuilder):
        """:param builder: The instruction builder containing sweeps."""
        self.sweeps: list[Sweep] = []
        self.sweep_names: list[str] = []  # list to preserve order
        self.sweep_sizes: list[int] = []
        self.device_assigns: list[DeviceUpdate] = []
        self.accessors: list[VariableAccessor] = []

        self._extract_sweeps(builder)
        self.total_combinations = prod(self.sweep_sizes)
        self._validate_device_assigns_variables(self.device_assigns)
        self._locate_variables(self.sweep_names)

    def create_flattened_builders(self) -> list[SweepInstance]:
        """Creates a list of builders with all combinations of sweep values injected.

        Returns a list of instances of the builder, containing the variable values, the
        builder with the injected values, and the device assignments to be applied before
        compiling the builders (if any).
        """

        builders = []
        for indices in product(*(range(size) for size in self.sweep_sizes)):
            variable_values = {}
            for i, index in enumerate(indices):
                sweep = self.sweeps[i]
                for key, val in sweep.variables.items():
                    variable_values[key] = val[index]
            new_builder = self._copy_and_inject_builder(variable_values)
            new_device_assigns = self._copy_and_inject_device_assigns(variable_values)
            builders.append(
                SweepInstance(
                    variable_values, new_builder, DeviceAssignSet(new_device_assigns)
                )
            )
        return builders

    def _create_new_builder(
        self, builder: QuantumInstructionBuilder, instructions: list[Instruction]
    ) -> QuantumInstructionBuilder:
        """Used to instantiate a new builder from an existing builder, defaulting to a
        QuantumInstructionBuilder if no hardware model is set within a builder."""
        new_builder = (
            builder.model.create_builder()
            if builder.model is not None
            else QuantumInstructionBuilder(hardware_model=None)
        )
        new_builder._instructions = instructions
        return new_builder

    def _extract_sweeps(self, builder: QuantumInstructionBuilder = None):
        """Extracts the sweeps from the instruction builder, removing them from the builder.

        A convinent side effect is that this also flattens any nested blocks in the builder,
        making it easier to inject variables.
        """

        instructions = []
        for instr in builder.instructions:
            if isinstance(instr, Sweep):
                self.sweeps.append(instr)
                self.sweep_names.extend(list(instr.variables.keys()))
                self.sweep_sizes.append(len(next(iter(instr.variables.values()))))
            elif isinstance(instr, DeviceUpdate):
                self.device_assigns.append(instr)
            else:
                instructions.append(instr)
        self._builder = self._create_new_builder(builder, instructions)

    def _locate_variables(self, names: list[str]):
        """Locate all variables in the instruction builder with the given names, returning
        a list of VariableAccesors that can be used to locate sweep variable instances.
        """

        self.accessors = []
        for i, instr in enumerate(self._builder._instructions):
            self.accessors.extend(VariableAccessor.create_accessors(instr, i, names))

    def _validate_device_assigns_variables(self, device_assigns: list[DeviceUpdate]):
        """Checks that any variables in device assignments corerspon"""
        for assign in device_assigns:
            if (
                isinstance(assign.value, Variable)
                and assign.value.name not in self.sweep_names
            ):
                raise ValueError(
                    f"Variable '{assign.value.name}' in device assignment is a variable "
                    "that is not allocated to any sweep."
                )

    def _copy_and_inject_builder(
        self, values: dict[str, complex | float]
    ) -> QuantumInstructionBuilder:
        """Injects the given values into the instructions using the injectors, and returns
        a copy of the builder with the modified instructions."""

        instructions_copy = [copy(instr) for instr in self._builder._instructions]
        builder = self._create_new_builder(self._builder, instructions_copy)

        for accessor in self.accessors:
            instr = builder._instructions[accessor.instruction_index]
            setattr(instr, accessor.attribute, values[accessor.variable_name])

        return builder

    def _copy_and_inject_device_assigns(
        self, values: dict[str, complex | float]
    ) -> list[DeviceUpdate]:
        """Injects the given values into the device assignments, returning a list of
        modified device assignments."""

        device_assigns = []
        for assign in self.device_assigns:
            new_assign = copy(assign)
            if isinstance((value := new_assign.value), Variable):
                new_assign.value = values[value.name]
            device_assigns.append(new_assign)
        return device_assigns



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from .fallthrough import FallthroughBackend as FallthroughBackend
from .waveform import PydWaveformBackend as PydWaveformBackend
from .waveform_v1.purr import WaveformV1Backend as WaveformV1Backend

DefaultBackend = WaveformV1Backend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod
from inspect import signature

from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.model.validators import MismatchingHardwareModelException
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class AbstractPipeline(ABC):
    """An abstraction of pipelines used in QAT to compile and execute quantum programs.

    Abstractly a pipeline needs to be given a hardware model and some target data that
    contains the parameters of the target device needed to correctly compile and execute.
    In the future, this might be relaxed so that the model and/or target data is not
    concretely needed to define a pipeline.
    """

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def model(self) -> QuantumHardwareModel | PhysicalHardwareModel: ...

    @property
    @abstractmethod
    def target_data(self) -> TargetData: ...

    def is_subtype_of(self, cls):
        """Matches a given type against the pipeline."""
        return isinstance(self, cls)


class BasePipeline(AbstractPipeline, ABC):
    """A base implementation of the abstract pipeline that provides access to the
    components of the pipeline.

    Subclasses should implement the `copy` and `copy_with_name` methods.
    """

    def __init__(self, name: str, model: QuantumHardwareModel, target_data: TargetData):
        self._name = name
        self._model = model
        self._target_data = target_data if target_data is not None else TargetData.default()

    @property
    def name(self) -> str:
        return self._name

    @property
    def model(self) -> QuantumHardwareModel:
        return self._model

    @property
    def target_data(self) -> TargetData:
        return self._target_data

    def copy(self) -> "BasePipeline":
        """Returns a new instance of the pipeline with the same components."""

        cls = self.__class__
        arg_names = signature(cls.__init__).parameters.keys()
        args = {k: getattr(self, k) for k in arg_names if k != "self"}
        return cls(**args)

    def copy_with_name(self, name: str) -> "BasePipeline":
        """Returns a new instance of the pipeline with the same components, but with a
        different name."""

        cls = self.__class__
        arg_names = signature(cls.__init__).parameters.keys()
        args = {k: getattr(self, k) for k in arg_names if k != "self"}
        args["name"] = name
        return cls(**args)

    def _validate_consistent_model(self, model: QuantumHardwareModel, *args):
        """Validates that the model is consistent across all components of the pipeline.

        :param model: The hardware model to validate against.
        :param args: The components of the pipeline to validate.
        """
        for component in args:
            if hasattr(component, "model") and component.model not in (model, None):
                raise MismatchingHardwareModelException(
                    f"{model} hardware does not match supplied hardware"
                )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import time
import uuid
from typing import List

from compiler_config.config import CompilerConfig
from qiskit import QuantumCircuit, qasm2, transpile
from qiskit.providers.basic_provider import BasicProviderJob, BasicSimulator
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from qat.purr.backends.realtime_chip_simulator import get_default_RTCS_hardware
from qat.purr.qat import execute_qasm
from qat.purr.utils.logger import get_default_logger

logger = get_default_logger()


def run_qasm_circuit(circuit):
    backend_sim = BasicSimulator()
    transpiled_qc = transpile(circuit, target=backend_sim.target)
    result = backend_sim.run(transpiled_qc).result()

    def _results_to_binary(results_dictionary):
        """
        Key value from Qiskit is in hex, aka ``0x2: 242``.
        Transform to binary, sort by actual integer value.
        """
        binary = {bin(int(key, 16))[2:]: val for key, val in results_dictionary.items()}
        fill_value = max([len(val) for val in binary.keys()])
        return {
            key.zfill(fill_value): val
            for key, val in sorted(binary.items(), key=lambda kvp: int(kvp[0]))
        }

    # If we're a single value, just return, otherwise return list of results (or None).
    if not any(result.results):
        return None
    elif len(result.results) == 1:
        return _results_to_binary(result.results[0].data.counts)
    else:
        return [_results_to_binary(res.counts) for res in result.results]


def run_qasm_from_file(qasm_file: str):
    """Runs the passed-in QASM file against the QASM state simulator."""
    return run_qasm_circuit(QuantumCircuit.from_qasm_file(qasm_file))


def run_qasm(qasm_str: str):
    """Runs the passed-in QASM string against the QASM state simulator."""
    return run_qasm_circuit(QuantumCircuit.from_qasm_str(qasm_str))


class QatBackend(BasicSimulator):
    """
    Basic qiskit backend built to run QASM on our own target machines.

    TODO: Expand this to become a proper back-end, as I don't believe we need to inherit
        off the QASM simulator as it stands.
    """

    def __init__(self, hardware=None, comp_config=None, **fields):
        super().__init__(**fields)
        self.hardware = hardware
        self.comp_config = comp_config

    def run(self, run_input, **run_options):
        if not isinstance(run_input, List):
            run_input = [run_input]

        if not all(isinstance(val, QuantumCircuit) for val in run_input):
            raise ValueError(
                "Experiment list contains non QuantumCircuit objects which we can't "
                "process at this time."
            )

        if self.comp_config is None:
            self.comp_config = CompilerConfig()
        qiskit_run_options = dict(self.options)
        for key, value in run_options.items():
            if key in qiskit_run_options:
                qiskit_run_options[key] = value
        self.comp_config.repeats = qiskit_run_options["shots"]
        self.comp_config.results_format.binary_count()

        experiment_info = [
            (
                qasm2.dumps(val),
                [creg.name for creg in val.cregs],
            )
            for val in run_input
        ]

        results = []
        full_start = time.time()
        status = "COMPLETED"
        try:
            for circ, creg_names in experiment_info:
                target_hardware = self.hardware or get_default_RTCS_hardware()
                start = time.time()
                results_data = execute_qasm(circ, target_hardware, self.comp_config)
                end = time.time()

                # TODO: Massage results data into a form that should be returned. As
                #   we're count it should be hex binary count representation, x01: 5024
                #   etc.

                # Below, solution for a single classical register. Not sure how qat
                # deals with multiple cregs.
                if len(creg_names) == 1:
                    creg_name = creg_names[0]
                    counts_qat = results_data[creg_name]
                    counts_qiskit = dict(
                        (hex(int(key[::-1], 2)), value)
                        for (key, value) in counts_qat.items()
                    )
                    results.append(
                        ExperimentResult(
                            self.comp_config.repeats,
                            True,
                            ExperimentResultData(counts=counts_qiskit),
                            status="DONE",
                            time_taken=end - start,
                        )
                    )

                else:
                    results.append(
                        ExperimentResult(
                            self.comp_config.repeats,
                            True,
                            ExperimentResultData(counts=results_data),
                            status="DONE",
                            time_taken=end - start,
                        )
                    )
        except Exception as ex:
            # Code further up infinitely recalls this method until it gets a result, so
            # catch and propagate message.
            status = str(ex)

        full_end = time.time()
        job_id = str(uuid.uuid4())
        return BasicProviderJob(
            self,
            job_id,
            Result(
                backend_name=self.name,
                backend_version=self.backend_version,
                job_id=job_id,
                results=results,
                status=status,
                success=any(results) and all(val.success for val in results),
                time_taken=full_end - full_start,
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

from copy import deepcopy

from pydantic import Field, field_validator, model_validator
from pydantic_extra_types.semantic_version import SemanticVersion
from semver import Version

from qat.model.device import Qubit, Resonator
from qat.model.error_mitigation import ErrorMitigation
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import (
    CalibratableUnitInterval,
    FrozenDict,
    FrozenSet,
    NoExtraFieldsModel,
    QubitCoupling,
    QubitId,
)
from qat.utils.uuid import uuid4

VERSION = Version(0, 0, 1)
log = get_default_logger()


class LogicalHardwareModel(NoExtraFieldsModel):
    """Models a hardware with a given connectivity.

    :param version: Semantic version of the hardware model.
    :param logical_connectivity: Connectivity of the qubits in the hardware model (directed graph).
    """

    version: SemanticVersion = Field(frozen=True, repr=False, default=VERSION)
    logical_connectivity: FrozenDict[QubitId, FrozenSet[QubitId]]

    calibration_id: str = Field(default_factory=lambda: str(uuid4()), frozen=True)

    @field_validator("version")
    def version_compatibility(version: Version):
        assert version.major == VERSION.major, (
            f"Direct instantiation requires major version compatibility (expected {VERSION.major}.Y.Z, found {version})"
        )
        assert VERSION >= version, (
            f"Latest supported hardware model version {VERSION}, found {version}"
        )
        return VERSION

    def __eq__(self, other: LogicalHardwareModel) -> bool:
        if type(self) is not type(other):
            return False

        if self.__class__.model_fields != other.__class__.model_fields:
            return False

        if self.version != other.version:
            return False

        if self.logical_connectivity != other.logical_connectivity:
            return False

        return True

    def __ne__(self, other: LogicalHardwareModel) -> bool:
        return not self.__eq__(other)


class PhysicalHardwareModel(LogicalHardwareModel):
    """Class for calibrating our QPU hardware.

    :param qubits: The superconducting qubits on the chip.
    :param physical_connectivity: The connectivities of the physical qubits on the QPU (undirected graph).
    :param logical_connectivity: The connectivities (directed graph) of the qubits used for compilation, which can be a subgraph of `physical_connectivity`.
    :param logical_connectivity_quality: Quality of the connections between the qubits.
    :param error_mitigation: Error mitigation strategy for this hardware model.
    """

    qubits: FrozenDict[QubitId, Qubit]
    physical_connectivity: FrozenDict[QubitId, FrozenSet[QubitId]] = Field(frozen=True)
    logical_connectivity: FrozenDict[QubitId, FrozenSet[QubitId]] | None = Field(
        default=None
    )
    logical_connectivity_quality: FrozenDict[QubitCoupling, CalibratableUnitInterval]
    error_mitigation: ErrorMitigation = ErrorMitigation()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._generate_mappings()

    def _generate_mappings(self):
        self._ids_to_pulse_channels = {}
        self._ids_to_physical_channels = {}
        self._pulse_channel_ids_to_physical_channel = {}
        self._pulse_channel_ids_to_device = {}
        self._qubits_to_qubit_ids = {}
        self._resonators_to_qubits = {}
        self._physical_channel_map = {}
        self._physical_channel_ids_to_qubit = {}
        self._physical_channel_ids_to_resonator = {}

        for qubit_id, qubit in self.qubits.items():
            self._qubits_to_qubit_ids[qubit] = qubit_id
            self._resonators_to_qubits[qubit.resonator] = qubit
            self._physical_channel_to_device_mapping(qubit)

            for device in [qubit, qubit.resonator]:
                phys_channel = device.physical_channel
                self._ids_to_physical_channels[phys_channel.uuid] = phys_channel
                self._physical_channel_map[phys_channel.name_index] = phys_channel

                for pulse_channel in device.all_pulse_channels:
                    self._ids_to_pulse_channels[pulse_channel.uuid] = pulse_channel
                    self._pulse_channel_ids_to_physical_channel[pulse_channel.uuid] = (
                        phys_channel
                    )
                    self._pulse_channel_ids_to_device[pulse_channel.uuid] = device

    def _physical_channel_to_device_mapping(self, qubit: Qubit):
        """
        populated the mapping from physical channel ids to devices (qubits and resonators)
        in the hardware model.
        The qubit mapping will also map the resonator physical channels to the qubit connected to
        the resonator.
        While the resonator mapping will only map the physical channels to the resonator.
        """
        self._physical_channel_ids_to_qubit.update(
            {
                qubit.physical_channel.uuid: qubit,
                qubit.resonator.physical_channel.uuid: qubit,
            }
        )

        self._physical_channel_ids_to_resonator.update(
            {qubit.resonator.physical_channel.uuid: qubit.resonator}
        )

    @model_validator(mode="before")
    def default_logical_connectivity(cls, data):
        if not data.get("logical_connectivity", None):
            logical_connectivity = deepcopy(data["physical_connectivity"])
            data["logical_connectivity"] = logical_connectivity

        if not data.get("logical_connectivity_quality", None):
            logical_connectivity_quality = {}

            for q1_index, connected_qubits in data["logical_connectivity"].items():
                for q2_index in connected_qubits:
                    logical_connectivity_quality[(q1_index, q2_index)] = 1.0

            data["logical_connectivity_quality"] = logical_connectivity_quality

        return data

    @field_validator("physical_connectivity")
    def validate_physical_connectivity_symmetry(cls, physical_connectivity):
        for q, connected_qs in physical_connectivity.items():
            for connected_q in connected_qs:
                if (
                    physical_connectivity.get(connected_q, None)
                    and q not in physical_connectivity[connected_q]
                ):
                    raise ValueError(
                        f"The topology is not symmetric, qubit {q} not present in connected qubits of {connected_q}."
                    )
        return physical_connectivity

    @model_validator(mode="after")
    def validate_connectivity(self):
        # Check if all qubits exist in physical connectivity.
        if len(self.qubits) > 1:  # 1Q-systems do not have any connectivity.
            assert self.qubits.keys() == self.physical_connectivity.keys(), (
                f"Inconsistent qubit ids for {self.qubits} and {self.physical_connectivity}."
            )

        # Check if logical connectivity is subset of physical connectivity.
        for qubit_index in self.logical_connectivity:
            if (
                self.logical_connectivity[qubit_index]
                and not self.logical_connectivity[qubit_index]
                <= self.physical_connectivity[qubit_index]
            ):
                raise ValueError(
                    "Logical connectivity must be a subgraph of the physical connectivity."
                )

        # Check if qubit cross resonance (cancellation) pulse channels agree with the physical connectivity.
        physical_connectivities = {
            (src, tgt) for (src, tgts) in self.physical_connectivity.items() for tgt in tgts
        }
        cross_resonance_edges = {
            (src, chan.auxiliary_qubit)
            for src, qubit in self.qubits.items()
            for chan in qubit.cross_resonance_pulse_channels.values()
        }
        cross_cancellation_edges = {
            (src, chan.auxiliary_qubit)
            for src, qubit in self.qubits.items()
            for chan in qubit.cross_resonance_cancellation_pulse_channels.values()
        }

        assert cross_resonance_edges == cross_cancellation_edges, (
            "Cross resonance channels mismatch cross resonance cancellation channels."
        )
        assert physical_connectivities == cross_resonance_edges, (
            "Cross resonance channels mismatch physical connectivity."
        )
        assert physical_connectivities == cross_cancellation_edges, (
            "Cross resonance cancellation channels mismatch physical connectivity."
        )

        # Check unique physical channel indices.
        physical_channel_indices = [
            qubit.physical_channel.name_index for qubit in self.quantum_devices
        ]
        assert len(physical_channel_indices) == len(set(physical_channel_indices)), (
            f"Physical channel indices must be unique, found {physical_channel_indices}."
        )

        return self

    @model_validator(mode="after")
    def validate_connectivity_quality(self):
        for q1_index, q2_index in self.logical_connectivity_quality:
            if not self.logical_connectivity.get(q1_index, None):
                raise IndexError(
                    f"The coupling ({q1_index}, {q2_index}) is not present in `logical_connectivity`."
                )

        for q1_index, connected_qubits in self.logical_connectivity.items():
            for q2_index in connected_qubits:
                if (q1_index, q2_index) not in self.logical_connectivity_quality:
                    raise IndexError(
                        f"The coupling ({q1_index}, {q2_index}) is not present in `logical_connectivity_quality`."
                    )

        return self

    @model_validator(mode="after")
    def validate_error_mitigation(self):
        if self.error_mitigation.is_enabled and not all(
            [q_id in self.qubits.keys() for q_id in self.error_mitigation.qubits]
        ):
            raise ValueError(
                "Please provide an error mitigation strategy for all qubits in the hardware model."
            )

        return self

    def __eq__(self, other: PhysicalHardwareModel) -> bool:
        if not super().__eq__(other):
            return False

        s_qubits = list(self.qubits.values())
        o_qubits = list(other.qubits.values())
        if len(s_qubits) != len(o_qubits):
            return False

        for s, o in zip(s_qubits, o_qubits):
            if s != o:
                return False

        return True

    @property
    def is_calibrated(self) -> bool:
        for qubit in self.qubits.values():
            if not qubit.is_calibrated:
                return False
        return True

    @property
    def number_of_qubits(self) -> int:
        return len(self.qubits)

    @property
    def physical_channel_map(self) -> dict:
        return self._physical_channel_map

    def qubit_with_index(self, index: int | QubitId) -> Qubit:
        return self.qubits[index]

    def index_of_qubit(self, qubit: Qubit) -> QubitId:
        """Returns the index of the given qubit in the hardware model."""
        return self._qubits_to_qubit_ids.get(qubit, None)

    def pulse_channel_with_id(self, id_: str):
        return self._ids_to_pulse_channels.get(id_, None)

    def physical_channel_with_id(self, id_: str):
        return self._ids_to_physical_channels.get(id_, None)

    def physical_channel_for_pulse_channel_id(self, id_: str):
        return self._pulse_channel_ids_to_physical_channel.get(id_, None)

    def device_for_pulse_channel_id(self, id_: str):
        return self._pulse_channel_ids_to_device.get(id_, None)

    def device_for_physical_channel_id(self, id_: str) -> Qubit | Resonator | None:
        if id_ in self._physical_channel_ids_to_resonator:
            return self._physical_channel_ids_to_resonator[id_]
        if id_ in self._physical_channel_ids_to_qubit:
            return self._physical_channel_ids_to_qubit[id_]
        log.warning(f"No device found for physical channel id {id_}.")

    def qubit_for_physical_channel_id(self, id_: str) -> Qubit | None:
        """Returns the qubit associated with the given physical channel id."""
        if id_ in self._physical_channel_ids_to_qubit:
            return self._physical_channel_ids_to_qubit[id_]
        log.warning(f"No qubit found for physical channel id {id_}.")

    def qubit_for_resonator(self, resonator: Resonator) -> Qubit | None:
        """Returns the qubit associated with the given resonator."""
        if resonator in self._resonators_to_qubits:
            return self._resonators_to_qubits[resonator]
        log.warning(f"No qubit found for resonator {resonator}.")

    @property
    def quantum_devices(self) -> list[Qubit, Resonator]:
        """
        Returns all quantum (an)harmonic oscillator devices
        in this hardware model as a list.
        """
        qubits = list(self.qubits.values())
        resonators = [qubit.resonator for qubit in qubits]
        return qubits + resonators

    def qubit_quality(self, physical_qubit_index: int):
        linear_mitigation = getattr(
            self.error_mitigation.readout_mitigation, "linear", None
        )
        if linear_mitigation is None:
            return 1.0
        readout_quality = linear_mitigation.get(physical_qubit_index, None)
        if readout_quality is not None:
            # TODO: COMPILER-706 linear readout mitigation currently a 2x2 matrix,
            #  we may want to change this to be a dictionary like in the old hardware model.
            return (readout_quality[0, 0] + readout_quality[1, 1]) / 2
        else:
            return 0


PydLogicalHardwareModel = LogicalHardwareModel
PydPhysicalHardwareModel = PhysicalHardwareModel
Model = PhysicalHardwareModel | QuantumHardwareModel



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.base import BaseBackend
from qat.engines import NativeEngine
from qat.frontend import BaseFrontend
from qat.middleend.base import BaseMiddleend
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.pipelines.base import AbstractPipeline
from qat.pipelines.pipeline import CompilePipeline, ExecutePipeline, Pipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.runtime.base import BaseRuntime


class CompilePipelineCache(CompilePipeline):
    """Accesses a cached full :class:`Pipeline`, allowing it to be recognised and treated as
    only a :class:`CompilePipeline`."""

    def __init__(self, pipeline: AbstractPipeline):
        if not pipeline.is_subtype_of(Pipeline):
            raise ValueError(
                f"Pipeline {pipeline.name} must be a full Pipeline to use the "
                "CompilePipelineCache wrapper."
            )
        self._pipeline = pipeline

    @property
    def name(self) -> str:
        return self._pipeline.name

    @property
    def model(self) -> QuantumHardwareModel | PhysicalHardwareModel:
        return self._pipeline.model

    @property
    def target_data(self) -> TargetData:
        return self._pipeline.target_data

    @property
    def frontend(self) -> BaseFrontend:
        return self._pipeline.frontend

    @property
    def middleend(self) -> BaseMiddleend:
        return self._pipeline.middleend

    @property
    def backend(self) -> BaseBackend:
        return self._pipeline.backend

    def copy_with_name(self, name):
        raise NotImplementedError(
            "CompilePipelineCache does not support copy_with_name. Use the original "
            "pipeline instead."
        )


class ExecutePipelineCache(ExecutePipeline):
    """Accesses a cached full :class:`Pipeline`, allowing it to be recognised and treated as
    only an :class:`ExecutePipeline`."""

    def __init__(self, pipeline: AbstractPipeline):
        if not pipeline.is_subtype_of(Pipeline):
            raise ValueError(
                f"Pipeline {pipeline.name} must be a full Pipeline to use the "
                "ExecutePipelineCache wrapper."
            )
        self._pipeline = pipeline

    @property
    def name(self) -> str:
        return self._pipeline.name

    @property
    def model(self) -> QuantumHardwareModel | PhysicalHardwareModel:
        return self._pipeline.model

    @property
    def target_data(self) -> TargetData:
        return self._pipeline.target_data

    @property
    def runtime(self) -> BaseRuntime:
        return self._pipeline.runtime

    @property
    def engine(self) -> NativeEngine:
        return self._pipeline.engine

    def copy_with_name(self, name):
        raise NotImplementedError(
            "ExecutePipelineCache does not support copy_with_name. Use the original "
            "pipeline instead."
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from typing import Dict, List, Optional

import numpy as np

from qat.core.config.configure import get_config
from qat.purr.backends.live_devices import (
    ControlHardware,
    ControlHardwareChannel,
    Instrument,
    LivePhysicalBaseband,
)
from qat.purr.backends.utilities import UPCONVERT_SIGN, get_axis_map
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    QubitCoupling,
    Resonator,
)
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.execution import QuantumExecutionEngine, SweepIterator
from qat.purr.compiler.hardware_models import ErrorMitigation, QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    Instruction,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
)
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()


def build_lucy_hardware(hw: QuantumHardwareModel):
    return apply_setup_to_hardware(hw, 8, 25e-9, 3.18e-07, 2.41e-06)


def apply_setup_to_hardware(
    hw: QuantumHardwareModel,
    qubit_count: int,
    pulse_hw_x_pi_2_width: float,
    pulse_hw_zx_pi_4_width: float,
    pulse_measure_width: float,
) -> QuantumHardwareModel:
    """Apply lucy to the passed-in hardware."""
    qubit_devices = []
    resonator_devices = []
    channel_index = 1

    control_hardware = ControlHardware()

    for primary_index in range(qubit_count):
        bb1 = LivePhysicalBaseband(
            f"LO{channel_index}", 5.5e9, if_frequency=250e6, instrument=control_hardware
        )
        bb2 = LivePhysicalBaseband(
            f"LO{channel_index + 1}",
            8.5e9,
            if_frequency=250e6,
            instrument=control_hardware,
        )
        hw.add_physical_baseband(bb1, bb2)

        ch1 = PhysicalChannel(f"CH{channel_index}", 1.0e-9, bb1, 1)
        ch2 = PhysicalChannel(
            f"CH{channel_index + 1}", 1.0e-9, bb2, 1, acquire_allowed=True
        )
        hw.add_physical_channel(ch1, ch2)

        resonator = Resonator(f"R{primary_index}", ch2)
        resonator.create_pulse_channel(ChannelType.measure, frequency=8.5e9)
        resonator.create_pulse_channel(ChannelType.acquire, frequency=8.5e9)

        # As our main system is a ring architecture we just attach every qubit in the
        # ring to the one on either side.
        # 2 has a connection to 1 and 3. This number wraps around, so we also have
        # 10-0-1 linkages.
        qubit = Qubit(primary_index, resonator, ch1)
        qubit.create_pulse_channel(ChannelType.drive, frequency=5.5e9)

        qubit.pulse_hw_x_pi_2.update({"width": pulse_hw_x_pi_2_width})
        qubit.pulse_hw_zx_pi_4.update({"width": pulse_hw_zx_pi_4_width})
        qubit.pulse_measure.update({"width": pulse_measure_width})

        qubit_devices.append(qubit)
        resonator_devices.append(resonator)
        channel_index = channel_index + 2

    # TODO: For backwards compatability cross resonance pulse channels are fully
    #   connected but coupled qubits are only in a ring architecture. I think it would be
    #   more appropriate for cross resonace channels to also be a ring architecture but
    #   that can be done in a later PR.
    for i, qubit in enumerate(qubit_devices):
        for other_qubit in qubit_devices:
            if qubit != other_qubit:
                qubit.create_pulse_channel(
                    auxiliary_devices=[other_qubit],
                    channel_type=ChannelType.cross_resonance,
                    frequency=5.5e9,
                    scale=50,
                )
                qubit.create_pulse_channel(
                    auxiliary_devices=[other_qubit],
                    channel_type=ChannelType.cross_resonance_cancellation,
                    frequency=5.5e9,
                    scale=0.0,
                )
            qubit.add_coupled_qubit(qubit_devices[(i + 1) % qubit_count])
            qubit.add_coupled_qubit(qubit_devices[(i - 1) % qubit_count])

    hw.add_quantum_device(*qubit_devices, *resonator_devices)
    hw.is_calibrated = True
    return hw


def sync_baseband_frequencies_to_value(hw: QuantumHardwareModel, lo_freq, target_qubits):
    # Round the drive channel frequencies to the multiple of 1kHz
    for qubit in target_qubits:
        hw.get_qubit(qubit).get_drive_channel().frequency = 1e9 * round(
            hw.get_qubit(qubit).get_drive_channel().frequency / 1e9, 6
        )

    for qubit in target_qubits:
        drive_channel = hw.get_qubit(qubit).get_drive_channel()
        physical_channel = drive_channel.physical_channel
        baseband = physical_channel.baseband
        log.info(f"Old baseband IF frequency of qubit '{qubit}': {baseband.if_frequency}")
        log.info(f"Old baseband frequency of qubit '{qubit}': {baseband.frequency}")
        baseband.if_frequency = float(round(drive_channel.frequency - lo_freq))
        baseband.frequency = lo_freq
        log.info(f"Baseband IF frequency of qubit '{qubit}' set to {baseband.if_frequency}")
        log.info(f"Baseband frequency of qubit '{qubit}' set to {baseband.frequency}")


class LiveHardwareModel(QuantumHardwareModel):
    def __init__(
        self,
        control_hardware: ControlHardware = None,
        shot_limit=10000,
        acquire_mode=None,
        repeat_count=1000,
        repetition_period=100e-6,
        error_mitigation: Optional[ErrorMitigation] = None,
        calibration_id: str = "",
    ):
        super().__init__(
            shot_limit=shot_limit,
            acquire_mode=acquire_mode or AcquireMode.INTEGRATOR,
            repeat_count=repeat_count,
            repetition_period=repetition_period,
            error_mitigation=error_mitigation,
            calibration_id=calibration_id,
        )
        self.control_hardware: ControlHardware = control_hardware
        self.instruments: Dict[str, Instrument] = {}
        self.qubit_direction_couplings: List[QubitCoupling] = []

    def create_engine(self, startup_engine: bool = True):
        return LiveDeviceEngine(self, startup_engine)

    def add_device(self, device):
        if isinstance(device, Instrument):
            self.add_instrument(device)
        else:
            super().add_device(device)

    def add_instrument(self, *instruments: Instrument):
        for instrument in instruments:
            if instrument.id in self.instruments:
                raise KeyError(f"Instrument with id '{instrument.id}' already exists.")
            self.instruments[instrument.id] = instrument

    def get_instrument(self, id_: str):
        return self.instruments.get(id_, None)

    def get_device(self, id_):
        if (device := self.get_instrument(id_)) is not None:
            return device
        return super().get_device(id_)

    def add_physical_baseband(self, *basebands: LivePhysicalBaseband):
        for baseband in basebands:
            if baseband.instrument.id not in self.instruments:
                self.add_instrument(baseband.instrument)
        super().add_physical_baseband(*basebands)

    def add_physical_channel(self, *physical_channels: ControlHardwareChannel):
        if self.control_hardware is not None:
            self.control_hardware.add_physical_channel(*physical_channels)
        super().add_physical_channel(*physical_channels)


class LiveDeviceEngine(QuantumExecutionEngine):
    """
    Backend that hooks up to our QPU's, currently hardcoded to particular fridges.
    This will only work when run on a machine physically connected to a QPU.
    """

    model: LiveHardwareModel

    def __init__(self, model: LiveHardwareModel, startup_engine: bool = True):
        super().__init__(model, startup_engine)

    def __enter__(self):
        return self

    def __del__(self):
        self.shutdown()

    def __exit__(self, exc, value, tb):
        self.shutdown()

    def startup(self):
        aggregate_result = []

        for instrument in self.model.instruments.values():
            instrument.connect()
            aggregate_result.append(instrument.is_connected)
        if self.model.control_hardware is not None:
            self.model.control_hardware.connect()
            aggregate_result.append(self.model.control_hardware.is_connected)
        for baseband in self.model.basebands.values():
            if isinstance(baseband, LivePhysicalBaseband):
                baseband.connect_to_instrument()
                aggregate_result.append(baseband.instrument.is_connected)
        for physical_channel in self.model.physical_channels:
            if isinstance(physical_channel, ControlHardwareChannel):
                for dc_bias_ch in physical_channel.dcbiaschannel_pair.values():
                    dc_bias_ch.connect_to_instrument()
                    aggregate_result.append(dc_bias_ch.instrument.is_connected)

        return all(aggregate_result)

    def shutdown(self):
        # Store the instrument connection statuses.
        # All instruments should be disconnected (represented as a False entry)
        aggregate_result = []
        for instrument in self.model.instruments.values():
            instrument.close()
            aggregate_result.append(instrument.is_connected)
        if self.model.control_hardware is not None:
            self.model.control_hardware.close()
            aggregate_result.append(self.model.control_hardware.is_connected)

        return not any(aggregate_result)

    def process_reset(self, position):
        raise NotImplementedError(
            "Active qubit reset is not currently implemented on live hardware."
        )

    def build_baseband_frequencies(
        self, pulse_channel_buffers: Dict[PulseChannel, np.ndarray]
    ):
        """Find fixed intermediate frequencies for physical channels if they exist."""
        baseband_freqs = {}
        baseband_freqs_fixed_if = {}
        for pulse_channel in pulse_channel_buffers.keys():
            if pulse_channel.fixed_if:
                if (
                    baseband_freqs_fixed_if.get(pulse_channel.physical_channel_id, False)
                    and baseband_freqs[pulse_channel.physical_channel_id]
                    != pulse_channel.frequency
                    - UPCONVERT_SIGN * pulse_channel.baseband_if_frequency
                ):
                    raise ValueError(
                        "Cannot fix the frequency for two pulse channels of different "
                        "frequencies on the same physical channel!"
                    )
                baseband_freqs[pulse_channel.physical_channel_id] = (
                    pulse_channel.frequency
                    - UPCONVERT_SIGN * pulse_channel.baseband_if_frequency
                )
                baseband_freqs_fixed_if[pulse_channel.physical_channel_id] = (
                    pulse_channel.fixed_if
                )
            else:
                if (
                    pulse_channel.physical_channel_id not in baseband_freqs_fixed_if
                    or not baseband_freqs_fixed_if[pulse_channel.physical_channel_id]
                ):
                    baseband_freqs_fixed_if[pulse_channel.physical_channel_id] = (
                        pulse_channel.fixed_if
                    )

        return baseband_freqs

    def _execute_on_hardware(
        self,
        sweep_iterator: SweepIterator,
        package: QatFile,
        interrupt: Interrupt = NullInterrupt(),
    ) -> Dict[str, np.ndarray]:
        if self.model.control_hardware is None:
            raise ValueError("Please add a control hardware first!")

        results = {}
        while not sweep_iterator.is_finished():
            sweep_iterator.do_sweep(package.instructions)
            metadata = {"sweep_iteration": sweep_iterator.get_current_sweep_iteration()}

            position_map = self.create_duration_timeline(package.instructions)
            pulse_channel_buffers = self.build_pulse_channel_buffers(position_map, True)
            buffers = self.build_physical_channel_buffers(pulse_channel_buffers)
            baseband_freqs = self.build_baseband_frequencies(pulse_channel_buffers)
            aq_map = self.build_acquire_list(position_map)

            data = {}
            for ch, buffer in buffers.items():
                if len(buffer) > 0:
                    data[ch] = buffer
                    if ch in baseband_freqs:
                        physical_channel = self.model.get_physical_channel(ch)
                        physical_channel.baseband.set_frequency(int(baseband_freqs[ch]))

            self.model.control_hardware.set_data(data)

            repetitions = package.repeat.repeat_count
            if (prt := package.repeat.passive_reset_time) is not None:
                repetition_time = self._calculate_repetition_time(prt, position_map)
            else:
                repetition_time = package.repeat.repetition_period

            for aqs in aq_map.values():
                if len(aqs) > 1:
                    raise ValueError(
                        "Multiple acquisitions are not supported on the same channel "
                        "in one sweep step."
                    )
                for aq in aqs:
                    physical_channel = aq.physical_channel
                    dt = physical_channel.sample_time
                    physical_channel.readout_start = aq.start * dt + (
                        aq.delay if aq.delay else 0.0
                    )
                    physical_channel.readout_length = aq.samples * dt
                    physical_channel.acquire_mode_integrator = (
                        aq.mode == AcquireMode.INTEGRATOR
                    )

            interrupt.if_triggered(metadata, throw=True)

            playback_results = self.model.control_hardware.start_playback(
                repetitions=repetitions, repetition_time=repetition_time
            )

            for channel, aqs in aq_map.items():
                if len(aqs) > 1:
                    raise ValueError(
                        "Multiple acquisitions are not supported on the same channel "
                        "in one sweep step."
                    )
                for aq in aqs:
                    response = playback_results[aq.physical_channel.id]
                    response_axis = get_axis_map(aq.mode, response)
                    for pp in package.get_pp_for_variable(aq.output_variable):
                        response, response_axis = self.run_post_processing(
                            pp, response, response_axis
                        )
                    var_result = results.setdefault(
                        aq.output_variable,
                        np.empty(
                            sweep_iterator.get_results_shape(response.shape),
                            response.dtype,
                        ),
                    )
                    sweep_iterator.insert_result_at_sweep_position(var_result, response)
        return results

    def _calculate_repetition_time(self, passive_reset_time, position_map) -> float:
        pc2samples = {pc: positions[-1].end for pc, positions in position_map.items()}
        durations = {pc: samples * pc.sample_time for pc, samples in pc2samples.items()}

        circuit_duration = max([duration for duration in durations.values()])

        return circuit_duration + passive_reset_time

    def optimize(self, instructions):
        with log_duration("Instructions optimized in {} seconds."):
            instructions = super().optimize(instructions)
            for pp in [val for val in instructions if isinstance(val, PostProcessing)]:
                if pp.acquire.mode == AcquireMode.SCOPE:
                    if (
                        pp.process == PostProcessType.MEAN
                        and ProcessAxis.SEQUENCE in pp.axes
                        and len(pp.axes) <= 1
                    ):
                        instructions.remove(pp)

                elif pp.acquire.mode == AcquireMode.INTEGRATOR:
                    if (
                        pp.process == PostProcessType.DOWN_CONVERT
                        and ProcessAxis.TIME in pp.axes
                        and len(pp.axes) <= 1
                    ):
                        instructions.remove(pp)

                    if (
                        pp.process == PostProcessType.MEAN
                        and ProcessAxis.TIME in pp.axes
                        and len(pp.axes) <= 1
                    ):
                        instructions.remove(pp)

            return instructions

    def validate(self, instructions: List[Instruction]):
        with log_duration("Instructions validated in {} seconds."):
            super().validate(instructions)

            consumed_qubits: List[str] = []
            chanbits_map = {}
            for inst in instructions:
                if isinstance(inst, PostProcessing):
                    if (
                        inst.acquire.mode == AcquireMode.SCOPE
                        and ProcessAxis.SEQUENCE in inst.axes
                    ):
                        raise ValueError(
                            "Invalid post-processing! Post-processing over SEQUENCE is "
                            "not possible after the result is returned from hardware "
                            "in SCOPE mode!"
                        )
                    elif (
                        inst.acquire.mode == AcquireMode.INTEGRATOR
                        and ProcessAxis.TIME in inst.axes
                    ):
                        raise ValueError(
                            "Invalid post-processing! Post-processing over TIME is not "
                            "possible after the result is returned from hardware in "
                            "INTEGRATOR mode!"
                        )
                    elif inst.acquire.mode == AcquireMode.RAW:
                        raise ValueError(
                            "Invalid acquire mode! The live hardware doesn't support "
                            "RAW acquire mode!"
                        )

                # Check if we've got a measure in the middle of the circuit somewhere.
                elif get_config().INSTRUCTION_VALIDATION.NO_MID_CIRCUIT_MEASUREMENT:
                    if isinstance(inst, Acquire):
                        for qbit in self.model.qubits:
                            if qbit.get_acquire_channel() == inst.channel:
                                consumed_qubits.append(qbit)
                    elif isinstance(inst, Pulse):
                        # Find target qubit from instruction and check whether it's been
                        # measured already.
                        acquired_qubits = [
                            (
                                (
                                    chanbits_map[chanbit]
                                    if chanbit in chanbits_map
                                    else chanbits_map.setdefault(
                                        chanbit,
                                        self.model._resolve_qb_pulse_channel(chanbit)[0],
                                    )
                                )
                                in consumed_qubits
                            )
                            for chanbit in inst.quantum_targets
                            if isinstance(chanbit, (Qubit, PulseChannel))
                        ]

                        if any(acquired_qubits):
                            raise ValueError(
                                "Mid-circuit measurements currently unable to be used."
                            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.qat import QATInput
from qat.utils.hardware_model import check_type_legacy_or_pydantic


class BaseFrontend(ABC):
    """
    Base class for frontend that scans a high-level language-specific, but target-agnostic, input
    :class:`QatInput` and verifies its syntax and semantics according to a specific source language
    (QASM, QIR, ...). The input is optimised so as to adhere to the underlying topology of the QPU
    and is then compiled to a target-agnostic intermediate representation (IR) :class:`QatIR` that
    can be further optimised in the middle end.

    Generally, frontends are picked to match the source language. For example, a QIR source
    program should be coupled with the :class:`QIRFrontend`. They implement compilation
    details that are specific to the source language via a pipeline, and parse it down to
    QAT IR.
    """

    def __init__(self, model: None | QuantumHardwareModel | PydHardwareModel = None):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = check_type_legacy_or_pydantic(model)

    @abstractmethod
    def emit(
        self,
        src: QATInput,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        """
        Compiles an input :class:`QatInput` down to :class:`QatIR` and emits it.
        :param src: The high-level input.
        :param res_mgr: Collection of analysis results with caching and aggregation
                        capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
                        compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        :return: An intermediate representation as an :class:`InstructionBuilder` which
                holds a list of instructions to be executed on the QPU.
        """
        ...

    @abstractmethod
    def check_and_return_source(self, src): ...

    @staticmethod
    def _check_metrics_and_config(res_mgr, met_mgr, compiler_config):
        if res_mgr is None:
            res_mgr = ResultManager()
        if met_mgr is None:
            met_mgr = MetricsManager()
        if compiler_config is None:
            compiler_config = CompilerConfig()

        return res_mgr, met_mgr, compiler_config



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict
from functools import reduce
from itertools import groupby
from math import prod

import numpy as np

from qat.backend.qblox.acquisition import Acquisition
from qat.executables import AcquireData
from qat.purr.compiler.instructions import AcquireMode


class ResultsAggregator:
    """Aggregates acquisition results from multiple acquisitions and programs.

    For now, we assume that the acquisitions always appear in some nested format, with
    iterations over variables being the outer-most levels in the nest. All of the different
    acquisition modes are supported:

    * RAW: The results for each iteration are saved as (number_shots, readout_length, )
    * SCOPE: The results for each iteration are saved as (readout_length, )
    * INTEGRATOR: The results for each iteration are saved as (number_shots, )

    The results for each iteration are the inner-most levels in the nest.

    Results are aggregated as a numpy array, and can be returned as either a numpy array or
    nested lists, with numpy arrays being the default. The nested lists is here if we need
    to replicated legacy behaviour.

    .. warning::

        The ResultsAggregator does not support batching of shots for SCOPE acquire modes.
    """

    def __init__(self):
        self._modes = {}
        self._shapes = {}
        self._results = {}
        self._indices = {}

    def _flatten_results_shape(
        self, mode: AcquireMode, shape: tuple[int, ...]
    ) -> tuple[int, ...]:
        """The outer dimensions of the results are treated as linear, allowing us to
        aggregate the results by just adding them to the single dimension. This is reshaped
        at the end to the correct shape."""

        if mode == AcquireMode.RAW or mode == AcquireMode.SCOPE:
            return (prod(shape[:-1]), shape[-1])
        return (prod(shape),)

    def clear(self):
        self._modes.clear()
        self._shapes.clear()
        self._results.clear()
        self._indices.clear()

    def append(self, playback: dict[str, np.ndarray], acquires: dict[str, AcquireData]):
        for name, data in playback.items():
            mode = self._modes.setdefault(name, acquires[name].mode)
            shape = self._shapes.setdefault(name, acquires[name].shape)
            result = self._results.setdefault(
                name,
                np.zeros(self._flatten_results_shape(mode, shape), dtype=np.complex128),
            )
            index = self._indices.setdefault(name, 0)

            data = data.reshape(self._flatten_results_shape(mode, data.shape))
            if data.shape[0] + index > result.shape[0]:
                raise ValueError(
                    "Attempting to append more results than allocated in the aggregator."
                )

            if mode == AcquireMode.RAW or mode == AcquireMode.SCOPE:
                if data.shape[1] != shape[-1]:
                    raise ValueError(
                        f"Expected readout length of {shape[-1]}, got {data.shape[1]}."
                    )
                result[index : index + data.shape[0], :] = data
            else:
                result[index : index + data.shape[0]] = data
            self._indices[name] = index + data.shape[0]

    def finalise(self, as_list: bool = False):
        results = {}
        for name, result in self._results.items():
            shape = self._shapes[name]
            finalised = np.reshape(result, shape)
            if as_list:
                finalised = finalised.tolist()
            results[name] = finalised
        return results


class QBloxAggregator:
    """
    Combines acquisition objects from multiple acquire instructions in multiple readout targets.
    Notice that :meth:`groupby` preserves (original) relative order, which makes it honour
    the (sequential) lexicographical order of the loop nest:

    playback[target]["acq_0"] contains (potentially) a list of acquisitions collected in the same
    order as the order in which the packages were sent to the FPGA.

    Although acquisition names are enough for unicity in practice, the playback's structure
    distinguishes different (multiple) acquisitions per readout target, thus making it more robust.
    """

    def __init__(self):
        self.playbacks: dict[str, list[Acquisition]] = defaultdict(list)

    def clear(self):
        self.playbacks.clear()

    def append(
        self, playback: dict[str, list[Acquisition]], acquires: dict[str, AcquireData]
    ):
        for pulse_channel_id, acquisitions in playback.items():
            self.playbacks[pulse_channel_id] += acquisitions

    def finalise(self) -> dict[str, dict[str, Acquisition]]:
        playback: dict[str, dict[str, Acquisition]] = {}
        for pulse_channel_id, acquisitions in self.playbacks.items():
            acquisitions.sort(key=lambda acquisition: acquisition.name)
            groups_by_name = groupby(acquisitions, lambda acquisition: acquisition.name)
            playback[pulse_channel_id] = {
                name: reduce(
                    lambda acq1, acq2: acq1 + acq2,
                    acqs,
                    Acquisition(),
                )
                for name, acqs in groups_by_name
            }

        return playback



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.frontend.parsers.qasm.qasm2 import (
    CloudQasmParser as CloudQasmParser,
    Qasm2Parser as Qasm2Parser,
    RestrictedQasm2Parser as RestrictedQasm2Parser,
)
from qat.frontend.parsers.qasm.qasm3 import Qasm3Parser as Qasm3Parser
from qat.frontend.parsers.qir import QIRParser as QIRParser



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC

from qat.model.loaders.base import BaseModelLoader
from qat.purr.compiler.hardware_models import LegacyHardwareModel


class BaseLegacyModelLoader(BaseModelLoader[LegacyHardwareModel], ABC): ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from bisect import insort
from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass, field
from itertools import chain

import numpy as np

from qat.backend.passes.purr.analysis import BindingResult, IterBound, TriageResult
from qat.backend.qblox.config.constants import Constants
from qat.core.pass_base import AnalysisPass
from qat.core.result_base import ResultInfoMixin, ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.instructions import (
    Acquire,
    Delay,
    DeviceUpdate,
    Instruction,
    Pulse,
    Variable,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxLegalisationPass(AnalysisPass):
    @staticmethod
    def phase_as_steps(phase_rad: float) -> int:
        """
        The instruction `set_ph_delta` expects the phase shift as a (potentially signed) integer operand.
        """

        phase_deg = np.rad2deg(phase_rad)
        phase_deg %= 360
        steps = int(round(phase_deg * Constants.NCO_PHASE_STEPS_PER_DEG))
        return steps

    @staticmethod
    def freq_as_steps(freq_hz: float) -> int:
        """
        The instruction `set_freq` expects the frequency as a (potentially signed) integer operand.
        """

        steps = int(round(freq_hz * Constants.NCO_FREQ_STEPS_PER_HZ))

        if (
            steps < -Constants.NCO_FREQ_LIMIT_STEPS
            or steps > Constants.NCO_FREQ_LIMIT_STEPS
        ):
            min_max_frequency_in_hz = (
                Constants.NCO_FREQ_LIMIT_STEPS / Constants.NCO_FREQ_STEPS_PER_HZ
            )
            raise ValueError(
                f"IF frequency must be in [-{min_max_frequency_in_hz:e}, {min_max_frequency_in_hz:e}] Hz. "
                f"Got {freq_hz:e} Hz"
            )

        return steps

    def amp_as_steps(self, amp: float) -> int:
        """
        The instruction `set_awg_offs` expects DAC ratio as a (potentially signed) integer operand.
        """

        amp_steps = int(amp.real * Constants.MAX_OFFSET)
        if amp_steps < Constants.MIN_OFFSET or amp_steps > Constants.MAX_OFFSET:
            raise ValueError(
                f"""
                Expected offset to be in range [{Constants.MIN_OFFSET}, {Constants.MAX_OFFSET}].
                Got {amp_steps} instead
                """
            )

        return amp_steps

    def _legalise_bound(self, name: str, bound: IterBound, inst: Instruction):
        legal_bound = bound

        if isinstance(inst, Acquire):
            num_bins = bound.count
            if num_bins > Constants.MAX_012_BINNED_ACQUISITIONS:
                raise ValueError(
                    f"""
                        Loop nest size would require {num_bins} acquisition memory bins which exceeds the maximum {Constants.MAX_012_BINNED_ACQUISITIONS}.
                        Please reduce number of points
                        """
                )
        elif isinstance(inst, Delay):
            if bound.start < Constants.GRID_TIME:
                log.warning(
                    f"Undefined runtime behaviour. Variable {name} has illegal lower bound"
                )
            if bound.end > Constants.MAX_WAIT_TIME:
                log.warning(
                    f"Undefined runtime behaviour. Will be batching variable {name} at runtime"
                )
        elif isinstance(inst, DeviceUpdate):
            if inst.attribute not in ["frequency", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {inst.attribute} for instruction {inst}"
                )

            if inst.attribute == "frequency":
                legal_bound = IterBound(
                    start=self.freq_as_steps(bound.start),
                    step=self.freq_as_steps(bound.step),
                    end=self.freq_as_steps(bound.end),
                    count=bound.count,
                )
            elif inst.attribute == "phase":
                legal_bound = IterBound(
                    start=self.phase_as_steps(bound.start),
                    step=self.phase_as_steps(bound.step),
                    end=self.phase_as_steps(bound.end),
                    count=bound.count,
                )
        else:
            attr2var = {
                attr: var
                for attr, var in vars(inst).items()
                if isinstance(var, Variable) and var.name == name
            }

            if attr2var and isinstance(inst, Pulse) and inst.shape != PulseShapeType.SQUARE:
                raise ValueError("Cannot process non-trivial pulses")
            if not attr2var:
                return legal_bound
            if len(attr2var) > 1:
                raise ValueError(
                    f"Unsafe analysis. Distinct attributes expecting the same variable bound {attr2var}"
                )
            attr, var = next(iter(attr2var.items()), (None, None))
            if attr not in ["width", "amp", "phase"]:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {attr} for instruction {inst}"
                )

            if attr == "width":
                if bound.start < Constants.GRID_TIME:
                    log.warning(
                        f"Undefined runtime behaviour. Variable {name} has illegal lower bound"
                    )
                if bound.end > Constants.MAX_WAIT_TIME:
                    log.warning(
                        f"Undefined runtime behaviour. Will be batching variable {name} at runtime"
                    )
            elif attr == "amp" and isinstance(inst, Pulse):
                legal_bound = IterBound(
                    start=self.amp_as_steps(bound.start),
                    step=self.amp_as_steps(bound.step),
                    end=self.amp_as_steps(bound.end),
                    count=bound.count,
                )
            elif attr == "phase":
                legal_bound = IterBound(
                    start=self.phase_as_steps(bound.start),
                    step=self.phase_as_steps(bound.step),
                    end=self.phase_as_steps(bound.end),
                    count=bound.count,
                )

        for attr, val in vars(legal_bound).items():
            if not isinstance(val, int):
                raise ValueError(
                    f"Illegal value {val} for attribute {attr}. Expected value to be an integer"
                )

        # Qblox registers are unsigned 32bit integers.
        return legal_bound.astype(np.uint32)

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        Performs target-dependent legalisation for QBlox.

            A) A repeat instruction with a very high repetition count is illegal because acquisition memory
        on a QBlox sequencer is limited. This requires optimal batching of the repeat instruction into maximally
        supported batches of smaller repeat counts.

        This pass does not do any batching. More features and adjustments will follow in future iterations.

            B) Previously processed variables such as frequencies, phases, and amplitudes still need digital conversion
        to a representation that's required by the QBlox ISA.

        + NCO's 1GHz frequency range by 4e9 steps:
            + [-500, 500] Mhz <=> [-2e9, 2e9] steps
            + 1 Hz            <=> 4 steps
        + NCO's 360° phase range by 1e9 steps:
            + 1e9    steps    <=> 2*pi rad
            + 125e6  steps    <=> pi/4 rad
        + Time and samples are quantified in nanoseconds
        + Amplitude depends on the type of the module:
            + [-1, 1]         <=> [-2.5, 2.5] V (for QCM)
            + [-1, 1]         <=> [-0.5, 0.5] V (for QRM)
        + AWG offset:
            + [-1, 1]         <=> [-32 768, 32 767]

        The last point is interesting as it requires knowledge of physical configuration of qubits and the modules
        they are wired to. This knowledge is typically found during execution and involving it early on would upset
        the rest of the compilation flow. In fact, it complicates this pass in particular, involves allocation
        concepts that should not be treated here, and promotes a monolithic compilation style. A temporary workaround
        is to simply assume the legality of amplitudes from the start whereby users are required to convert
        the desired voltages to the equivalent ratio AOT.

        This pass performs target-dependent conversion as described in part (B). More features and adjustments
        will follow in future iterations.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target in triage_result.target_map:
            rw_result = binding_result.rw_results[target]
            bound_result = binding_result.iter_bound_results[target]
            legal_bound_result: dict[str, IterBound] = deepcopy(bound_result)

            qblox_bounds: dict[str, set[IterBound]] = defaultdict(set)
            for name, instructions in rw_result.reads.items():
                for inst in instructions:
                    legal_bound = self._legalise_bound(name, bound_result[name], inst)
                    qblox_bounds[name].add(legal_bound)

            for name, bound in bound_result.items():
                if name in qblox_bounds:
                    bound_set = qblox_bounds[name]
                    if len(bound_set) > 1:
                        raise ValueError(
                            f"Ambiguous Qblox bounds for variable {name} in target {target}"
                        )
                    legal_bound_result[name] = next(iter(bound_set))

            # TODO: the proper way is to produce a new result and invalidate the old one
            binding_result.iter_bound_results[target] = legal_bound_result
        return ir


@dataclass
class AllocationManager:
    _reg_pool: list[str] = field(
        default_factory=lambda: sorted(
            f"R{index}" for index in range(Constants.NUMBER_OF_REGISTERS)
        )
    )
    _lbl_counters: dict[str, int] = field(default_factory=dict)

    registers: dict[str, str] = field(default_factory=dict)
    labels: dict[str, str] = field(default_factory=dict)

    def reg_alloc(self, name: str) -> str:
        if name in self.registers:
            log.warning(f"Returning a register already allocated for Variable {name}")
            register = self.registers[name]
        elif len(self._reg_pool) < 1:
            raise IndexError(
                "Out of registers. Attempting to use more registers than available in the Q1 sequence processor"
            )
        else:
            register = self._reg_pool.pop(0)
            self.registers[name] = register

        return register

    def reg_free(self, register: str) -> None:
        if register in self._reg_pool:
            raise RuntimeError(f"Cannot free register '{register}' as it's not in use")
        insort(self._reg_pool, register)
        self.registers = {
            var: reg for var, reg in self.registers.items() if reg != register
        }

    @contextmanager
    def reg_borrow(self, name: str):
        """
        Short-lived register allocation
        """

        register = self.reg_alloc(name)
        yield register
        self.reg_free(register)

    def label_gen(self, name: str):
        counter = self._lbl_counters.setdefault(name, 0)
        self._lbl_counters[name] += 1
        label = f"block_{name}_{counter}"
        self.labels[name] = label
        return label


@dataclass
class PreCodegenResult(ResultInfoMixin):
    alloc_mgrs: dict[PulseChannel, AllocationManager] = field(
        default_factory=lambda: defaultdict(lambda: AllocationManager())
    )


class PreCodegenPass(AnalysisPass):
    """
    Precedes code generation. The context-based emitter needs registers pre-allocated for every
    live variable in program. This pass performs a naive register allocation through a manager object.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)
        result = PreCodegenResult()

        for target in triage_result.target_map:
            alloc_mgr = result.alloc_mgrs[target]
            iter_bound_result = binding_result.iter_bound_results[target]
            reads = binding_result.rw_results[target].reads
            writes = binding_result.rw_results[target].writes

            alloc_mgr.reg_alloc("zero")
            names = set(chain(*[iter_bound_result.keys(), reads.keys(), writes.keys()]))
            for name in names:
                alloc_mgr.reg_alloc(name)
                alloc_mgr.label_gen(name)

        res_mgr.add(result)

        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
def get_default_pipelines():
    """Returns a list of descriptions for default full pipelines, compile pipelines, and
    execute pipelines."""

    from qat.core.config.descriptions import PipelineInstanceDescription

    from .echo import echo8, echo16, echo32

    full_pipelines = [
        PipelineInstanceDescription(name=echo8.name, pipeline=echo8),
        PipelineInstanceDescription(name=echo16.name, pipeline=echo16),
        PipelineInstanceDescription(name=echo32.name, pipeline=echo32, default=True),
    ]
    compile_pipelines = []
    execute_pipelines = []
    return full_pipelines, compile_pipelines, execute_pipelines



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from warnings import warn

from qat.backend.waveform_v1.purr.codegen import WaveformV1Backend
from qat.frontend import AutoFrontend
from qat.middleend import DefaultMiddleend
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class WaveformV1CompilePipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`WaveformV1Backend`.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs. Please
        select an appropriate execution pipeline if you wish to run the compiled programs.
    """

    def __init__(self, *args, **kwargs):
        warn(
            "WaveformV1 support is deprecated and will be removed in v4.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> CompilePipeline:
        if engine is not None:
            log.warning(
                "An engine was provided to the WaveformV1CompilationPipeline, but it will "
                "be ignored. "
            )

        target_data = target_data if target_data is not None else TargetData.default()
        return CompilePipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend.default_for_purr(model),
            middleend=DefaultMiddleend(model, target_data),
            backend=WaveformV1Backend(model),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from importlib.metadata import version
from warnings import warn

from compiler_config.config import CompilerConfig as CompilerConfig

from qat.core.config.configure import get_config
from qat.core.config.session import QatSessionConfig as QatSessionConfig
from qat.core.qat import QAT as QAT
from qat.pipelines.pipeline import Pipeline as Pipeline
from qat.purr.qat import (
    execute as execute,
    execute_qasm as execute_qasm,
    execute_qasm_with_metrics as execute_qasm_with_metrics,
    execute_qir as execute_qir,
    execute_qir_with_metrics as execute_qir_with_metrics,
    execute_with_metrics as execute_with_metrics,
    fetch_frontend as fetch_frontend,
)
from qat.purr.qatconfig import QatConfig

qatconfig: QatConfig = get_config()

# TODO: Remove version checking, COMPILER-658
if version("qiskit") < "2.0.0":
    warn(
        "Support for Qiskit <2.0.0 is deprecated in QAT and Qiskit >=2.0.0 will soon "
        "be required. Please update your local Qiskit dependency to >=2.0.0.",
        DeprecationWarning,
        stacklevel=2,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import abstractmethod

from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class ModelUpdateChecker:
    """A mixin for model loaders that adds an interface for checking if models are
    up-to-date for models that are expected to evolve within its lifetime."""

    @abstractmethod
    def is_up_to_date(self, model: PhysicalHardwareModel | QuantumHardwareModel) -> bool:
        """Used to check if the model is up-to-date. To be implemented by subclasses."""
        ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import re
from collections import defaultdict
from copy import deepcopy

import numpy as np

from qat.ir.waveforms import SofterSquareWaveform, SoftSquareWaveform, SquareWaveform
from qat.model.device import (
    AcquirePulseChannel,
    CalibratableAcquire,
    CalibratablePulse,
    CrossResonanceCancellationPulseChannel,
    CrossResonancePulseChannel,
    DrivePulseChannel,
    FreqShiftPulseChannel,
    MeasurePulseChannel,
    PhysicalBaseband,
    Qubit,
    QubitPhysicalChannel,
    QubitPulseChannels,
    Resonator,
    ResonatorPhysicalChannel,
    ResonatorPulseChannels,
    SecondStatePulseChannel,
)
from qat.model.error_mitigation import (
    CalibratableUnitInterval2x2Array,
    ErrorMitigation,
    ReadoutMitigation,
)
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.devices import ChannelType, PulseChannelView, PulseShapeType
from qat.purr.compiler.instructions import CustomPulse
from qat.utils.pydantic import FrozenDict
from qat.utils.uuid import uuid_randomiser

number_mask = re.compile("[0-9]+")


def get_number_from_string(s: str) -> int | None:
    number_match = number_mask.search(s)
    if number_match is not None:
        return int(number_match.group())
    return None


def convert_purr_echo_hw_to_pydantic(legacy_hw, seed_uuid: bool = True):
    new_qubits = {}
    logical_connectivity = defaultdict(set)

    if seed_uuid:
        uuid_randomiser.seed(legacy_hw.calibration_id)

    for qubit in legacy_hw.qubits:
        # Add topology of qubit. Since the topology is not always stored both in
        # the `qubit_direction_couplings`, look into `coupled_qubits` as well.
        if couplings := getattr(legacy_hw, "qubit_direction_couplings", None):
            coupled_q_indices = [
                c.direction[1] for c in couplings if c.direction[0] == qubit.index
            ]
        else:
            coupled_q_indices = [q.index for q in qubit.coupled_qubits]
        logical_connectivity[qubit.index].update(coupled_q_indices)

        # Physical baseband
        phys_bb_q = qubit.physical_channel.baseband
        new_phys_bb_q = PhysicalBaseband(
            frequency=phys_bb_q.frequency, if_frequency=phys_bb_q.if_frequency
        )

        phys_bb_r = qubit.measure_device.physical_channel.baseband
        new_phys_bb_r = PhysicalBaseband(
            frequency=phys_bb_r.frequency, if_frequency=phys_bb_r.if_frequency
        )

        # Physical channel
        phys_channel_q = qubit.physical_channel
        new_phys_ch_q = QubitPhysicalChannel(
            baseband=new_phys_bb_q,
            block_size=phys_channel_q.block_size,
            name_index=get_number_from_string(phys_channel_q.id),
        )

        phys_channel_r = qubit.measure_device.physical_channel
        new_phys_ch_r = ResonatorPhysicalChannel(
            baseband=new_phys_bb_r,
            block_size=phys_channel_r.block_size,
            swap_readout_iq=getattr(phys_channel_r, "swap_readout_IQ", False),
            name_index=get_number_from_string(phys_channel_r.id),
        )

        # Resonator
        measure_pulse_channel = qubit.measure_device.get_pulse_channel(ChannelType.measure)
        pulse_measure = deepcopy(qubit.pulse_measure)
        pulse_measure.pop("shape")
        new_measure_pulse_channel = MeasurePulseChannel(
            frequency=measure_pulse_channel.frequency,
            imbalance=measure_pulse_channel.imbalance,
            scale=_process_real_or_complex(measure_pulse_channel.scale),
            phase_iq_offset=measure_pulse_channel.phase_offset,
            pulse=CalibratablePulse(waveform_type=SquareWaveform, **pulse_measure),
        )

        meas_acq = deepcopy(qubit.measure_acquire)
        if meas_acq.get("weights") is None:
            meas_acq["weights"] = []
        elif isinstance(meas_acq.get("weights"), CustomPulse):
            meas_acq["weights"] = meas_acq["weights"].samples

        acquire_pulse_channel = qubit.measure_device.get_pulse_channel(ChannelType.acquire)
        new_acquire_pulse_channel = AcquirePulseChannel(
            frequency=acquire_pulse_channel.frequency,
            imbalance=acquire_pulse_channel.imbalance,
            scale=_process_real_or_complex(acquire_pulse_channel.scale),
            phase_iq_offset=acquire_pulse_channel.phase_offset,
            acquire=CalibratableAcquire(**meas_acq),
        )

        new_res_pulse_channels = ResonatorPulseChannels(
            measure=new_measure_pulse_channel, acquire=new_acquire_pulse_channel
        )

        resonator = Resonator(
            physical_channel=new_phys_ch_r,
            pulse_channels=new_res_pulse_channels,
        )

        # Qubit pulse channels
        drive_pulse_channel = qubit.get_pulse_channel(ChannelType.drive)
        pulse_hw_x_pi_2 = deepcopy(qubit.pulse_hw_x_pi_2)
        pulse_hw_x_pi_2.pop("shape")
        pulse_hw_x_pi = getattr(qubit, "pulse_hw_x_pi", None)
        if pulse_hw_x_pi is not None:
            pulse_hw_x_pi = deepcopy(pulse_hw_x_pi)
            pulse_hw_x_pi.pop("shape")
            pulse_hw_x_pi = CalibratablePulse(**pulse_hw_x_pi)

        new_drive_pulse_channel = DrivePulseChannel(
            frequency=drive_pulse_channel.frequency,
            imbalance=drive_pulse_channel.imbalance,
            scale=_process_real_or_complex(drive_pulse_channel.scale),
            phase_iq_offset=drive_pulse_channel.phase_offset,
            pulse=CalibratablePulse(**pulse_hw_x_pi_2),
            pulse_x_pi=pulse_hw_x_pi,
        )

        try:
            freqshift_pulse_channel = qubit.get_pulse_channel(ChannelType.freq_shift)
            new_freqshift_pulse_channel = FreqShiftPulseChannel(
                frequency=freqshift_pulse_channel.frequency,
                imbalance=freqshift_pulse_channel.imbalance,
                scale=_process_real_or_complex(freqshift_pulse_channel.scale),
                phase_iq_offset=freqshift_pulse_channel.phase_offset,
                active=freqshift_pulse_channel.active,
                amp=freqshift_pulse_channel.amp,
                phase=getattr(freqshift_pulse_channel.pulse_channel, "phase", 0.0),
            )
        except KeyError:
            new_freqshift_pulse_channel = FreqShiftPulseChannel()

        try:
            secondstate_pulse_channel = qubit.get_pulse_channel(ChannelType.second_state)
            new_secondstate_pulse_channel = SecondStatePulseChannel(
                frequency=secondstate_pulse_channel.frequency,
                imbalance=secondstate_pulse_channel.imbalance,
                scale=_process_real_or_complex(secondstate_pulse_channel.scale),
                phase_iq_offset=secondstate_pulse_channel.phase_offset,
            )
        except KeyError:
            new_secondstate_pulse_channel = SecondStatePulseChannel()

        new_cross_resonance_pulse_channels = {}
        new_cross_resonance_cancellation_pulse_channels = {}

        for pulse_channel in qubit.pulse_channels.values():
            if (
                isinstance(pulse_channel, PulseChannelView)
                and pulse_channel.auxiliary_devices
            ):
                aux_qubit = pulse_channel.auxiliary_devices[0].index

                if pulse_channel.channel_type == ChannelType.cross_resonance:
                    zx_pi_4 = deepcopy(qubit.pulse_hw_zx_pi_4.get(f"Q{aux_qubit}", None))

                    waveform_type = (
                        SofterSquareWaveform
                        if zx_pi_4["shape"] == PulseShapeType.SOFTER_SQUARE
                        else SoftSquareWaveform
                    )
                    zx_pi_4.pop("shape")

                    new_cr_pulse_channel = CrossResonancePulseChannel(
                        auxiliary_qubit=aux_qubit,
                        frequency=pulse_channel.frequency,
                        imbalance=pulse_channel.imbalance,
                        scale=_process_real_or_complex(pulse_channel.scale),
                        phase_iq_offset=pulse_channel.phase_offset,
                        zx_pi_4_pulse=CalibratablePulse(
                            waveform_type=waveform_type, **zx_pi_4
                        ),
                    )
                    new_cross_resonance_pulse_channels[aux_qubit] = new_cr_pulse_channel

                elif pulse_channel.channel_type == ChannelType.cross_resonance_cancellation:
                    new_crc_pulse_channel = CrossResonanceCancellationPulseChannel(
                        auxiliary_qubit=aux_qubit,
                        frequency=pulse_channel.frequency,
                        imbalance=pulse_channel.imbalance,
                        scale=_process_real_or_complex(pulse_channel.scale),
                        phase_iq_offset=pulse_channel.phase_offset,
                    )
                    new_cross_resonance_cancellation_pulse_channels[aux_qubit] = (
                        new_crc_pulse_channel
                    )

        new_cross_resonance_pulse_channels = FrozenDict(new_cross_resonance_pulse_channels)

        new_cross_resonance_cancellation_pulse_channels = FrozenDict(
            new_cross_resonance_cancellation_pulse_channels
        )

        new_qubit_pulse_channels = QubitPulseChannels(
            drive=new_drive_pulse_channel,
            freq_shift=new_freqshift_pulse_channel,
            second_state=new_secondstate_pulse_channel,
            cross_resonance_channels=new_cross_resonance_pulse_channels,
            cross_resonance_cancellation_channels=new_cross_resonance_cancellation_pulse_channels,
        )

        new_qubit = Qubit(
            physical_channel=new_phys_ch_q,
            pulse_channels=new_qubit_pulse_channels,
            resonator=resonator,
            mean_z_map_args=qubit.mean_z_map_args,
            discriminator=qubit.discriminator[0],
            direct_x_pi=getattr(qubit, "direct_x_pi", False),
        )
        new_qubits[qubit.index] = new_qubit

    new_qubits = FrozenDict(new_qubits)
    physical_connectivity = defaultdict(set)
    for q1_index in logical_connectivity:
        for q2_index in logical_connectivity[q1_index]:
            physical_connectivity[q1_index].add(q2_index)
            physical_connectivity[q2_index].add(q1_index)

    if legacy_hw.qubit_direction_couplings:
        logical_connectivity_quality = {}
        for coupling in legacy_hw.qubit_direction_couplings:
            logical_connectivity_quality[coupling.direction] = (
                coupling.quality / 100.0
            )  # We assume that the quality is in [0, 100] in the legacy hw model.
    else:
        logical_connectivity_quality = None

    # Convert Error Mitigation if present
    if (error_mitigation := legacy_hw.error_mitigation) is not None and (
        readout_mitigation := getattr(
            legacy_hw.error_mitigation, "readout_mitigation", None
        )
    ) is not None:
        linear = {}
        if readout_mitigation.linear:
            for qubit_idx, qubit_map in readout_mitigation.linear.items():
                linear[qubit_idx] = CalibratableUnitInterval2x2Array(
                    np.array(
                        [
                            [qubit_map["0|0"], qubit_map["0|1"]],
                            [qubit_map["1|0"], qubit_map["1|1"]],
                        ]
                    )
                )
        linear = FrozenDict(linear)

        readout_mitigation = ReadoutMitigation(
            linear=linear,
            matrix=getattr(readout_mitigation, "matrix", None),
            m3_available=getattr(readout_mitigation, "m3_available", False),
        )
        error_mitigation = ErrorMitigation(readout_mitigation=readout_mitigation)
    else:
        error_mitigation = ErrorMitigation()

    calibration_id = getattr(legacy_hw, "calibration_id", "")
    new_hw = PhysicalHardwareModel(
        logical_connectivity=logical_connectivity,
        logical_connectivity_quality=logical_connectivity_quality,
        physical_connectivity=physical_connectivity,
        qubits=new_qubits,
        calibration_id=calibration_id,
        error_mitigation=error_mitigation,
    )

    if seed_uuid:
        uuid_randomiser.seed()

    return new_hw


def _process_real_or_complex(value: float | complex) -> float | complex:
    if isinstance(value, complex) and value.imag == 0.0:
        return value.real

    return value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from uuid import uuid4

from pydantic import BaseModel, Field


class PulseChannel(BaseModel):
    """An IR-level representation of a pulse channel.

    This holds the properties of a pulse channel that are relevant at compile-time, and
    decouples it from the hardware model, allowing dynamic representation and creation of
    hardware models.

    Eventually, this might be replaced with a more formal IR operation the declares pulse
    channels a symbols, with lookups defer to symbol tables.
    """

    uuid: str = Field(default_factory=lambda: str(uuid4()), frozen=True)
    frequency: float
    imbalance: float = 1.0
    phase_iq_offset: float = 0.0
    scale: float | complex = 1.0 + 0.0j
    physical_channel_id: str

    def __hash__(self):
        return hash(self.uuid)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import numpy as np

from qat.core.pass_base import ValidationPass
from qat.ir.lowered import PartitionedIR
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import Acquire, CustomPulse, Pulse
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class NCOFrequencyVariability(ValidationPass):
    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        model = next((a for a in args if isinstance(a, QuantumHardwareModel)), None)

        if not model:
            model = kwargs.get("model", None)

        if not model or not isinstance(model, QuantumHardwareModel):
            raise ValueError(
                f"Expected to find an instance of {QuantumHardwareModel} in arguments "
                f"list, but got {model} instead"
            )

        for channel in model.pulse_channels.values():
            if channel.fixed_if:
                raise ValueError("Cannot allow constance of the NCO frequency")
        return ir


class NoAcquireWeightsValidation(ValidationPass):
    """Some target machines do not support :class:`Acquire` instructions that contain weights.
    This pass can be used to validate that this is the case."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        has_filters = [inst.filter for inst in ir.instructions if isinstance(inst, Acquire)]
        if any(has_filters):
            raise NotImplementedError(
                "Acquire filters are not implemented for this target machine."
            )
        return ir


class NoMultipleAcquiresValidation(ValidationPass):
    """Some target machines do not support multiple :class:`Acquire` instructions on the
    same channel. This validation pass should be used to verify this."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        physical_channels = [
            inst.channel.physical_channel_id
            for inst in ir.instructions
            if isinstance(inst, Acquire)
        ]
        if len(physical_channels) != len(set(physical_channels)):
            raise NotImplementedError(
                "Multiple acquisitions on a single channel is not supported for this target machine."
            )
        return ir


class NoAcquiresWithDifferentWeightsValidation(ValidationPass):
    """Some target machines do not support multiple :class:`Acquire` instructions with different
    filters on the same pulse channel. This validation pass should be used to verify this.
    """

    def run(self, ir: PartitionedIR, *args, **kwargs):
        for pulse_ch, acquires in ir.acquire_map.items():
            samples = []
            for acquire in acquires:
                if isinstance(acquire.filter, CustomPulse):
                    samples.append(acquire.filter.samples)
                elif isinstance(acquire, Pulse):
                    raise TypeError(
                        "Type of the acquire filter can only be `CustomPulse` or `None`."
                    )

            samples = np.array(samples)
            if not np.all(samples == samples[0]):
                raise ValueError(
                    f"Cannot have multiple `Acquire`s on the same pulse channel ({pulse_ch.full_id()}) with different weights."
                )

        return ir


# TODO - bring in stuff from verification.py in here in the form of a pass (or a bunch of passes)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import re
from typing import Any

from pydantic import Field
from pydantic_extra_types.semantic_version import SemanticVersion
from semver import Version

from qat.ir.waveforms import Waveform
from qat.model.device import Resonator
from qat.model.hardware_model import PhysicalHardwareModel
from qat.utils.pydantic import (
    CalibratablePositiveFloat,
    NoExtraFieldsFrozenModel,
    NoExtraFieldsModel,
    find_all_subclasses,
)
from qat.utils.units import Frequency, Scale, Time

VERSION = Version(0, 0, 1)


class FeatureMetadata(NoExtraFieldsFrozenModel):
    """
    Exposed metadata for various languages, integrations and features.
    """

    name: str
    description: str
    version: SemanticVersion = Field(frozen=True, repr=False, default=VERSION)
    enabled: bool = True


class OpenPulseConstraints(NoExtraFieldsModel):
    pulse_control_contraints: str = None
    max_scale: float = 1.0
    max_waveform_amplitude: float = 1.0
    awg_frequency_bandwidth: Frequency = Frequency(amount=1, scale=Scale.DEFAULT)
    max_pulse_duration: Time = Time(amount=120, scale=Scale.MICRO)
    min_pulse_duration: Time = Time(amount=8, scale=Scale.NANO)
    min_pulse_time: Time = Time(amount=8, scale=Scale.NANO)


class OpenPulseFrame(NoExtraFieldsModel):
    qubits: list[str]
    port_id: str
    frequency: CalibratablePositiveFloat
    bandwidth_centre: CalibratablePositiveFloat
    phase: float = 0.0


class OpenPulsePort(NoExtraFieldsModel):
    direction: str
    type: str
    associated_qubits: list[int]


class OpenPulseFeatures(FeatureMetadata):
    """
    Features specific to OpenPulse.
    """

    ports: dict[str, OpenPulsePort]
    frames: dict[str, Any]
    waveforms: dict[str, Any]
    constraints: OpenPulseConstraints | None = None

    @staticmethod
    def from_hardware(model: PhysicalHardwareModel) -> OpenPulseFeatures:
        if not isinstance(model, PhysicalHardwareModel):
            raise ValueError(
                "OpenPulseFeatures can only be generated from a `model` of type `PhysicalHardwareModel`."
            )

        # Generate frames from the hardware model.
        frames = {}
        for qubit_id, qubit in model.qubits.items():
            for pulse_channel in qubit.all_qubit_and_resonator_pulse_channels:
                pulse_ch_name = pulse_channel.__class__.__name__
                ignored = any(
                    ignored_pc in pulse_ch_name.lower()
                    for ignored_pc in ["freqshift", "secondstate"]
                )

                if not ignored:
                    pulse_ch_type = re.sub(r"PulseChannel$", "", pulse_ch_name)
                    pulse_ch_type = CamelCase_to_snake(pulse_ch_type)

                    qubits = [f"Q{qubit_id}"]
                    device = model.device_for_pulse_channel_id(pulse_channel.uuid)

                    # Generate a unique frame name for each pulse channel.
                    frame_name = (
                        f"r{qubit_id}_"
                        if isinstance(device, Resonator)
                        else f"q{qubit_id}_"
                    )

                    if hasattr(
                        pulse_channel, "auxiliary_qubit"
                    ):  # CR and CRC pulse channels
                        frame_name += f"q{pulse_channel.auxiliary_qubit}_"
                        qubits.append(f"Q{pulse_channel.auxiliary_qubit}")

                    frame_name += pulse_ch_type

                    frames[frame_name] = OpenPulseFrame(
                        qubits=qubits,
                        port_id=f"channel_{device.physical_channel.name_index}",
                        frequency=pulse_channel.frequency,
                        bandwidth_centre=device.physical_channel.baseband.frequency,
                        phase=0.0,
                    )

        # Generate ports from the hardware model.
        ports = {}
        for qubit_id, qubit in model.qubits.items():
            port_name_q = f"channel_{qubit.physical_channel.name_index}"
            port_name_r = f"channel_{qubit.resonator.physical_channel.name_index}"
            ports[port_name_q] = OpenPulsePort(
                direction="one-way",
                type="port_type_1",
                associated_qubits=[qubit_id],
            )
            ports[port_name_r] = OpenPulsePort(
                direction="two-way",
                type="port_type_1",
                associated_qubits=[qubit_id],
            )

        # Generate waveforms from the available waveform classes.
        waveforms = find_all_subclasses(Waveform)
        waveforms = {CamelCase_to_snake(waveform.name()): "test" for waveform in waveforms}

        constraints = OpenPulseConstraints()

        return OpenPulseFeatures(
            name="OpenPulse",
            description="Features for OpenPulse integration.",
            version=VERSION,
            ports=ports,
            frames=frames,
            waveforms=waveforms,
            constraints=constraints,
        )

    def to_json_dict(self):
        # This method is to keep backwards compatibility with purr OpenPulse features.
        blob = self.model_dump()
        return {"open_pulse": blob}


def CamelCase_to_snake(name: str) -> str:
    # Insert underscore before each capital letter (except the first), then lowercase
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()
